 # si.py
# A class representing the SI prefixes (SIPrefix) and a list of the SI
# base units (si_unit_stems): length (L), mass (M), time (T), temperature
# (Theta), amount of substance (Q), current (C) and luminous intensity (I).
#
# Copyright (C) 2012-2016 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

class SIPrefix(object):
    """ A little class describing SI prefixes. """
    def __init__(self, prefix, name, power):
        self.prefix = prefix
        self.name = name
        self.power = power
        self.fac = 10**power

# Here are the SI prefixes that we recognise.
si_prefixes = { 'y': SIPrefix('y', 'yocto', -24),
                'z': SIPrefix('z', 'zepto', -21),
                'a': SIPrefix('a', 'atto', -18),
                'f': SIPrefix('f', 'femto', -15),
                'p': SIPrefix('p', 'pico', -12),
                'n': SIPrefix('n', 'nano', -9),
                'μ': SIPrefix('μ', 'micro', -6),
                'm': SIPrefix('m', 'milli', -3),
                'c': SIPrefix('c', 'centi', -2),
                'd': SIPrefix('d', 'deci', -1),
                'k': SIPrefix('k', 'kilo', 3),
                'M': SIPrefix('M', 'mega', 6),
                'G': SIPrefix('G', 'giga', 9),
                'T': SIPrefix('T', 'tera', 12),
                'P': SIPrefix('P', 'peta', 15),
                'E': SIPrefix('E', 'exa', 18),
                'Z': SIPrefix('Z', 'zetta', 21),
                'Y': SIPrefix('Y', 'yotta', 24),
              }

# The base SI unit stems for length, time, mass, amount of substance,
# thermodynamic temperature, luminous intenstiy and current respectively:
si_unit_stems = ('m', 's', 'g', 'mol', 'K', 'cd', 'A')




# list_base_units.py
import sys
from base_unit import base_units

def list_units_text():
    for group_name, base_unit_group in base_units:
        print()
        print(group_name)
        print('='*len(group_name))
        print('{0:6s} {1:30s} {2:20s}'.format('Unit', 'Name', 'Dimensions'))
        print('{0:6s} {1:30s} {2:20s}'.format('----', '----', '----------'))
        for base_unit in base_unit_group:
            print('{0:6s} {1:30s} {2:20s}'.format(unicode(base_unit.stem),
                base_unit.name, base_unit.dims))

def list_units_html():
    for group_name, base_unit_group in base_units:
        print('<h4>{0}</h4>'.format(group_name))
        print('<table class="my-table">')
        print('<tr><th>Unit</th><th>Name</th><th>Dimensions</th></tr>')
        for base_unit in base_unit_group:
            # NB we need to encode('utf-8') to pipe in that encoding to the
            # shell, e.g. to create a file
            print('<tr><td>{0:s}</td><td>{1:s}</td><td>{2:s}</td><td></tr>'
                    .format(unicode(base_unit.stem), base_unit.name,
                            base_unit.dims).encode('utf-8'))
        print('</table>')
    

try:
    list_type = sys.argv[1]
    assert list_type in ('text', 'html')
except (IndexError, AssertionError):
    print('usage:\n{0} <list_type>'.format(sys.argv[0]))
    print('where <list_type> is text or html')
    sys.exit(1)

if list_type == 'text':
    list_units_text()
else:
    list_units_html()



# symbol.py
# A class representing a symbol (perhaps the label for a physical quantity
# represented as a Quantity object) with a name in text,  LaTeX and HTML.
#
# Copyright (C) 2012-2016 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

class Symbol(object):
    """
    A Python class representing a symbol - typically the label for a physical
    quantity - with plain-text (utf8-encoded string), LaTeX, and HTML versions
    and perhaps a definition or description.

    """

    def __init__(self, name=None, latex=None, html=None, definition=None):
        """
        Initialize the Symbol object: if they are not given, the LaTeX and HTML
        descriptions of the Symbol are set to its plain-text name.

        """

        self.name = name
        self.latex = latex
        if not latex:
            self.latex = name
        self.html = html
        if not html:
            self.html = name
        self.definition = definition

    def __str__(self):
        """
        Return the string representation of the Symbol, its name attribute.
        """
        return self.name



from .symbol import Symbol
import numpy as np

class qnArrayError(Exception):
    def __init__(self, error_str):
        self.error_str = error_str
    def __str__(self):
        return self.error_str

class qnArray(Symbol):
    def __init__(self, name=None, latex=None, html=None, values=None,
                 units=None, sd=None, definition=None):
        Symbol.__init__(self, name, latex, html, definition)
        if type(values)==list:
            self.values = np.array(values)
        elif type(values)==numpy.ndarray:
            self.values = values
        else:
            raise qnArrayError



# quantity.py
# A class representing physical quantity, with name, units and uncertainty.
#
# Copyright (C) 2012-2016 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

import re
import math
#import numpy as np
from .symbol import Symbol
from .units import Units, UnitsError

class QuantityError(Exception):
    """
    An Exception class for errors that might occur whilst manipulating
    Quantity objects.

    """
    def __init__(self, error_str):
        self.error_str = error_str
    def __str__(self):
        return self.error_str

class Quantity(Symbol):
    """
    A Python class representing a physical quantity. This class extends the
    Symbol class, so provides a means of naming the quantity using plain text
    (a utf8-encoded string), LaTeX and HTML, and a definition string.
    A Quantity also has a value, units, and an absolute uncertainty, expressed
    as a standard deviation (sd).

    """

    def __init__(self, name=None, latex=None, html=None, value=None,
                 units=None, sd=None, definition=None):
        """
        Initialize the Quantity object: set up its name as for the base,
        Symbol class, and set the quantity's value, sd and units (which
        can be None, a string representing the units, or a Units object and
        will be converted to a Units object in any case).

        """

        Symbol.__init__(self, name, latex, html, definition)
        self.value = value
        self.sd = sd
        if units is None:
            self.units = Units([])
        else:
            self.units = Units(units)

    def __str__(self):
        """ A simple string representation of the Quantity. """

        if self.name:
            return '%s = %s %s' % (self.name, self.value, self.units)
        else:
            return '%s %s' % (self.value, self.units)
            
    __repr__ = __str__

    def value_as_str(self, nsd_digits=2, small=1.e-3, large=1.e5):
        """
        Return a string representation of the parameter and its
        standard deviation in the conventional format used in
        the spectroscopic literature:
        Examples:
        0.04857623(71) for 4.857623e-2 +/- 7.1e-7
        -0.412(40) for -0.412 +/- 0.04
        1.7324(38)e7 for 17324000 +/- 380
        
        Notes:
        This routine has not been rigorously tested, because I wrote
        it at 1am. Therefore, it would be a good idea to output the
        raw parameter values and standard deviations somewhere as
        well as using the output from this routine...
        
        Arguments:
        nsd_digits: the number of digits in the standard deviation
            (the default is 2)
        small: parameter values smaller than small are output in
            scientific notation (<mantissa>e<power>)
        large: parameter values larger than large are output in
            scientific notation (<mantissa>e-<power>)
        
        """
        N, sd = self.value, self.sd
        
        if not sd:
            return str(N)

        absN = abs(N)
        if (absN < small or absN > large) and absN != 0.:
            # scientific notation
            power = int(math.floor(math.log(absN, 10)))
            N *= pow(10, -power)
            sd *= pow(10, -power)
            dp = int(-math.log(sd,10))+nsd_digits
            if dp < 0:
                dp = 0
            sd_digits = int(round(sd * pow(10, dp)))
            fmt = "%." + str(dp) + "f(" + str(sd_digits) + ")e%d"
            return fmt % (N, power)
        else:
            # regular floating point notation
            dp = int(-math.log(sd, 10)) + nsd_digits
            if dp < 0:
                dp = 0
            sd_digits = int(round(sd * pow(10, dp)))
            fmt = "%." + str(dp) + "f(" + str(sd_digits) + ')'

        return fmt % N

    def as_str(self, b_name=True, b_sd=True, b_units=True):
        """
        Output the Quantity object as a string, optionally with name,
        standard deviation and units.

        """

        s = []
        if b_name and self.name:
            s.append('%s = ' % self.name)
        if b_sd:
            s.append(self.value_as_str(b_sd))
        else:
            s.append(str(self.value))
        if b_units and self.units.has_units():
            s.append(' %s' % str(self.units))
        return ''.join(s)

    def convert_units_to(self, new_units, force=None):
        """
        Convert the Quantity from one set of units to an equivalent set,
        complaining loudly if the two sets of units do not have the same
        dimensions.

        """

        if self.value is None:
            return
        to_units = Units(new_units)
        fac = self.units.conversion(to_units, force)

#        self.value *= fac
#        if self.sd is not None:
#            self.sd *= fac
#        self.units = to_units
        if self.sd is not None:
            return Quantity(value = self.value*fac, units = new_units, sd = self.sd*fac)
        else:
            return Quantity(value = self.value*fac, units = new_units)

#    def draw_from_dist(self, shape=None):
#        """
#        Return a value or number array of values drawn from the normal
#        distribution described by this Quantity's mean and standard
#        deviation. shape is the shape of the NumPy array to return, or
#        None (the default) to return a single scalar value from the
#        distribution.
#
#        """
#
#        if self.sd is None:
#            raise ValueError('Quantity instance {} has no defined standard'
#                             ' deviation.'.format(self.name))
#
#        return np.random.normal(loc=self.value, scale=self.sd, size=shape)

    def __add__(self, other):
        """
        Add two Quantity objects; they must have the same units. Errors
        are propagated, but assumed to be uncorrelated.

        """

        if type(other) != Quantity:
            raise TypeError
        if self.value is None or other.value is None:
            raise ValueError
        if self.units != other.units:
            raise UnitsError('Can\'t add two quantities with different'
                             ' units: %s and %s' % (self.units, other.units))
        if self.sd is None or other.sd is None:
            sd = None
        else:
            sd = math.hypot(self.sd, other.sd)
        return Quantity(value=self.value+other.value, units=self.units, sd=sd)

    def __sub__(self, other):
        """
        Subtract two Quantity objects; they must have the same units. Errors
        are propagated, but assumed to be uncorrelated.

        """

        if type(other) != Quantity:
            raise TypeError
        if self.value is None or other.value is None:
            raise ValueError
        if self.units != other.units:
            raise UnitsError('Can\'t subtract two quantities with different'
                             ' units: %s and %s' % (self.units, other.units))
        if self.sd is None or other.sd is None:
            sd = None
        else:
            sd = math.hypot(self.sd, other.sd)
        return Quantity(value=self.value-other.value, units=self.units, sd=sd)

    def __mul__(self, other):
        """
        Multiply this quantity by a number or another quantity. Errors are
        propagated, but assumed to be uncorrelated.

        """

        if self.value is None:
            raise ValueError
        if type(other) in (int, float):
            if self.sd is None:
                sd = None
            else:
                sd = abs(other) * self.sd
            return Quantity(value=self.value*other, units=self.unit, sd=sd)
        else:
            if type(other) != Quantity:
                raise TypeError
            if other.value is None:
                raise ValueError
            value = self.value * other.value
            if not self.sd or not other.sd:
                sd = None
            else:
                sd = value * math.hypot(self.sd/self.value,
                                        other.sd/other.value)
            units = self.units * other.units
            return Quantity(value=value, units=units, sd=sd)

    def __truediv__(self, other):
        """
        Divide this quantity by a number or another quantity. Errors are
        propagated, but assumed to be uncorrelated.

        """

        if self.value is None:
            raise ValueError
        if type(other) in (int, float):
            if self.sd is None:
                sd = None
            else:
                sd = abs(other) / self.sd
            return Quantity(value=self.value/other, units=self.units, sd=sd)
        else:
            if type(other) != Quantity:
                raise TypeError
            if other.value is None:
                raise ValueError
            value = self.value / other.value
            if not self.sd or not other.sd:
                sd = None
            else:
                sd = value * math.hypot(self.sd/self.value,
                                        other.sd/other.value)
            units = self.units / other.units
            return Quantity(value=value, units=units, sd=sd)

    def __rtruediv__(self, other):
        return self.__truediv__(other)
    
    def __pow__(self, power):
        return Quantity(value = self.value**power, 
                        units = self.units**power)

    @classmethod
    def parse(self, s_quantity, name=None, units=None, sd=None,
              definition=None):
        s_quantity = s_quantity.strip()
        if '=' in s_quantity:
            fields = s_quantity.split('=')
            s_name = fields[0].strip()
            if s_name:
                name = s_name
            s_quantity = fields[1].strip()
        fields = s_quantity.split(' ')
        s_valsd = fields[0]
        if len(fields) == 2:
            units = Units(fields[1])
        # force lower case, and replace Fortran-style 'D'/'d' exponents
        s_valsd = s_valsd.lower()
        s_valsd = s_valsd.replace('d','e')
        if 'e' in s_valsd:
            s_mantsd, s_exp = s_valsd.split('e')
            exp = int(s_exp)
        else:
            s_mantsd = s_valsd
            exp = 0
        patt = r'([+-]?\d*\.?\d*)\(?(\d+)?\)?'
        m = re.match(patt, s_mantsd)
        if not m:
            raise QuantityError('Failed to parse string into quantity:\n'\
                                '%s' % s_mantsd)
        s_mantissa, s_sd = m.groups()
        mantissa = float(s_mantissa)
        value = mantissa * 10**exp
        sd = None
        if s_sd:
            if '.' in s_mantissa:
                ndp = len(s_mantissa) - s_mantissa.index('.') - 1
            else:
                ndp = 0
            sd = float(s_sd) * 10**(exp-ndp)
        return Quantity(name=name, value=value, units=units, sd=sd)



# dimensions.py
# A class representing the dimensions of a physical quantity's units, in
# terms of powers of length (L), mass (M), time (T), temperature (Theta),
# amount of substance (Q), current (C) and luminous intensity (I).
#
# Copyright (C) 2012-2016 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

class Dimensions(object):
    # these are the abbreviations for Length, Mass, Time, Temperature,
    # Quantity (amount of substance), Current, and Luminous Intensity:
    dim_names = ['L', 'M', 'T', 'Theta', 'Q', 'C', 'I']
    dim_desc = ['length', 'mass', 'time', 'temperature', 'amount',
                'current', 'luminous intensity']
    dim_index = {}
    for i, dim_name in enumerate(dim_names):
        dim_index[dim_name] = i

    def __init__(self, dims=None, **kwargs):
        self.dims = [0]*7
        if dims:
            # initialize by dims array
            if not kwargs:
                self.dims = dims
            else:
                print('bad initialisation of Dimensions object')
                sys.exit(1)
        else:
            # initialize by keyword arguments
            for dim_name in kwargs:
                self.dims[self.dim_index[dim_name]] = kwargs[dim_name]

    def __mul__(self, other):
        new_dims = []
        for i, dim in enumerate(self.dims):
            new_dims.append(dim + other.dims[i])
        return Dimensions(tuple(new_dims))

    def __truediv__(self, other):
        new_dims = []
        for i, dim in enumerate(self.dims):
            new_dims.append(dim - other.dims[i])
        return Dimensions(tuple(new_dims))

    def __pow__(self, pow):
        new_dims = []
        for dim in self.dims:
            new_dims.append(dim * pow)
        return Dimensions(tuple(new_dims))

    def __str__(self):
        s_dims = []
        for i, dim_name in enumerate(self.dim_names):
            if self.dims[i] != 0:
                this_s_dim = dim_name
                if self.dims[i] != 1:
                    this_s_dim += '%d' % self.dims[i]
                s_dims.append(this_s_dim)
        if len(s_dims) == 0:
            return '[dimensionless]'
        return '.'.join(s_dims)

    def __repr__(self):
        return str(self.dims)

    def __eq__(self, other):
        for i, dim in enumerate(self.dims):
            if other.dims[i] != dim:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

d_dimensionless = Dimensions()
d_quantity = Dimensions(Q=1)
d_frequency = Dimensions(T=-1)
d_length = Dimensions(L=1)
d_area = d_length**2
d_volume = d_length**3
d_time = Dimensions(T=1)
d_mass = Dimensions(M=1)
d_energy = Dimensions(M=1, L=2, T=-2)
d_force = d_mass * d_length / d_time**2
d_pressure = d_force / d_area
d_current = Dimensions(C=1)
d_charge = d_current * Dimensions(T=1)
d_voltage = d_energy / d_charge     # 1 V = 1 J/C
d_magfield_strength = d_voltage * d_time / d_area   # 1 T = 1 V.s/m^2
d_magnetic_flux = d_voltage * d_time    # 1 Wb = 1 V.s
d_temperature = Dimensions(Theta=1)


# atom_unit.py
# A class for, and attributes used in the parsing of a single "atom unit"
# identified by a base unit, SI-prefix and exponent, such as 'kg', 's-1',
# 'nm3', 'GHz'.
#
# Copyright (C) 2012 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

import sys
from pyparsing import Word, Group, Literal, Suppress, ParseException, oneOf,\
                      Optional
from .si import si_prefixes
from .base_unit import BaseUnit, base_unit_stems
from .dimensions import Dimensions

# pyparsing stuff for parsing unit strings:
caps = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
lowers = caps.lower()
letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_0'
digits = '123456789'
exponent = Word(digits + '-')
prefix = oneOf(list(si_prefixes.keys()))
ustem = Word(letters + 'Å' + 'Ω')
uatom = (Group( '1' | (Optional(prefix) + ustem)) + Optional(exponent))\
            | (Group( '1' | ustem) + Optional(exponent))

# floating point equality and its negation, to some suitable tolerance
def feq(f1, f2, tol=1.e-10):
    return abs(f1-f2) <= tol
def fneq(f1, f2, tol=1.e-10):
    return not feq(f1, f2, tol)

class UnitsError(Exception):
    """
    An Exception class for errors that might occur whilst manipulating units.

    """
    def __init__(self, error_str):
        self.error_str = error_str
    def __str__(self):
        return self.error_str

class AtomUnit(object):
    """
    AtomUnit is a class to represent a single BaseUnit, possibly with an SI
    prefix and possibly raised to some power.

    """

    def __init__(self, prefix, base_unit, exponent=1):
        """ Initialize the AtomUnit object. """

        self.base_unit = base_unit
        self.exponent = exponent
        self.dims = self.base_unit.dims ** self.exponent

        # get the SI prefix (if present), and its 'factor' (10 raised to its
        # the power it represents
        self.si_fac = 1.
        self.prefix = prefix
        if prefix is not None:
            try:
                self.si_prefix = si_prefixes[prefix]
            except KeyError:
                raise UnitsError('Invalid or unsupported SI prefix: %s'
                                                        % prefix)
            self.si_fac = self.si_prefix.fac
        # now calculate the factor relating this AtomUnit to its
        # corresponding SI unit:
        self.si_fac = (self.si_fac * self.base_unit.fac) ** self.exponent

    @classmethod
    def parse(self, s_unit_atom):
        """
        Parse the string s_unit_atom into an AtomUnit object and return
        it.

        """

        try:
            uatom_data = uatom.parseString(s_unit_atom)
        except ParseException:
            raise
            raise UnitsError('Invalid unit atom syntax: %s' % s_unit_atom)

        # uatom_data comes back as (([prefix], <stem>), [exponent])
        if len(uatom_data[0]) == 1:
            # no prefix, just the stem
            prefix = None
            stem = uatom_data[0][0]
        else:
            prefix = uatom_data[0][0]
            stem = uatom_data[0][1]
            # we need to ensure that units such as 'mmHg' get resolved
            # properly (ie to mmHg, not to milli-mHg):
            if stem not in base_unit_stems:
                prefix = None
                stem = ''.join(uatom_data[0])
        try:
            base_unit = base_unit_stems[stem]
        except:
            raise UnitsError("Unrecognised unit: %s" % s_unit_atom)

        # if there is an exponent, determine what it is (default is 1)
        exponent = 1
        if len(uatom_data) == 2:
            exponent = int(uatom_data[1])
        return AtomUnit(prefix, base_unit, exponent)

    def __pow__(self, power):
        """ Return the current AtomUnit raised to a specified power. """
        return AtomUnit(self.prefix, self.base_unit, self.exponent * power)

    def __str__(self):
        """ String representation of this AtomUnit. """
        s = ''
        if self.prefix:
            s = self.prefix
        s_exponent = ''
        if self.exponent != 1:
            s_exponent = str(self.exponent)
        return ''.join([s, str(self.base_unit), s_exponent])

    def __repr__(self):
        return str(self)

    def prefix_base_eq(self, other):
        """
        Compare two AtomUnits and return true if they have the same BaseUnit
        and the same prefix.

        """

        if self.prefix == other.prefix and self.base_unit == other.base_unit:
            return True
        return False




# base_unit.py
# A class representing a "base unit", identified as a single unit with no
# prefix or exponent, such as 'g', 'hr', 'bar', 's'.
#
# Copyright (C) 2012-2016 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

from .dimensions import *

class BaseUnit(object):
    """
    A BaseUnit is a commonly-used single unit without a prefix, for example:
    m (metre, length), erg (energy), bar (pressure). BaseUnit objects have
    a description and know their dimensions in terms of powers of Length,
    Time, Mass, etc. (described by a Dimensions object). 

    """

    def __init__(self, stem, name, unit_type, fac, description, latex,
                 dims=None):
        """ Initialize the BaseUnit object. """
        # BaseUnit stem, e.g. 'm', 'g', 'bar', ...
        self.stem = stem
        # BaseUnit name, e.g. 'metre', 'gram', 'bar', ...
        self.name = name
        # BaseUnit type, e.g. 'length', 'mass', 'pressure', ...
        self.unit_type = unit_type
        # fac is the relationship between this base unit and the SI base
        # unit, e.g. m: 1. (m is SI), g: 1.e-3 (1 g = 1.e-3 kg),
        # bar: 1.e5 (1 bar = 10e5 Pa)
        self.fac = fac
        # any further description of the BaseUnit
        self.description = description
        # LaTex representation of the BaseUnit
        self.latex = latex
        # BaseUnit dimensions, as a Dimension object
        self.dims = dims

    def __eq__(self, other):
        if self.stem == other.stem:
            return True

    def __str__(self):
        """ String representation of the BaseUnit is just its 'stem'. """
        return self.stem

base_units = (
('Unity', [
BaseUnit('1', 'unity', 'unity', 1., '', '1', d_dimensionless),
]),

('SI unit stems', [
BaseUnit('m', 'metre', 'length', 1., '', 'm', d_length),
BaseUnit('s', 'second', 'time', 1., '', 's', d_time),
BaseUnit('g', 'gram', 'mass', 1.e-3, '', 'g', d_mass),
BaseUnit('K', 'kelvin', 'temperature', 1., '', 'K', Dimensions(Theta=1)),
BaseUnit('mol', 'mole', 'amount', 1., '', 'mol', Dimensions(Q=1)),
BaseUnit('A', 'amp', 'current', 1., '', 'A', Dimensions(C=1)),
BaseUnit('cd', 'candela', 'luminous intensity', 1., '', 'cd', Dimensions(I=1))
]),

('Derived SI units', [
BaseUnit('N', 'newton', 'force', 1., '', 'N', d_force),
BaseUnit('J', 'joule', 'energy', 1., '', 'J', d_energy),
BaseUnit('W', 'watt', 'power', 1., '', 'W', d_energy / d_time),
BaseUnit('Pa', 'pascal', 'pressure', 1., '', 'Pa', d_pressure),
BaseUnit('C', 'coulomb', 'charge', 1., '', 'C', d_charge),
BaseUnit('V', 'volt', 'voltage', 1., '', 'V', d_voltage),
BaseUnit('Ω', 'ohm', 'electric resistance', 1, '', r'\Omega',
         d_voltage / d_current),
BaseUnit('F', 'farad', 'capacitance', 1., '', 'F', d_charge / d_voltage),
BaseUnit('Wb', 'weber', 'magnetic flux', 1, '', 'Wb', d_magnetic_flux),
BaseUnit('H', 'henry', 'inductance', 1, '', 'H', d_magnetic_flux / d_current),
BaseUnit('S', 'siemens', 'electric conductance', 1, '', 'S',
         d_current / d_voltage),
BaseUnit('T', 'tesla', 'magnetic field strength', 1., '', 'T',
         d_magfield_strength),
BaseUnit('Hz', 'hertz', 'cyclic frequency', 1., '', 'Hz', d_time**-1),
BaseUnit('rad', 'radian', 'angle', 1., '', 'rad', d_dimensionless),
BaseUnit('sr', 'steradian', 'solid angle', 1., '', 'sr', d_dimensionless),
]),

('Non-SI pressure units', [
BaseUnit('bar', 'bar', 'pressure', 1.e5, '', 'bar', d_pressure),
BaseUnit('atm', 'atmosphere', 'pressure', 1.01325e5, '', 'atm', d_pressure),
BaseUnit('Torr', 'torr', 'pressure', 133.322368, '', 'Torr', d_pressure),
# (but see e.g. Wikipedia for the precise relationship between Torr and mmHg
BaseUnit('mmHg', 'millimetres of mercury', 'pressure', 133.322368, '', 'mmHg',
         d_pressure),
]),

('cgs units', [
BaseUnit('dyn', 'dyne', 'force', 1.e-5, '', 'dyn', d_force),
BaseUnit('erg', 'erg', 'energy', 1.e-7, '', 'erg', d_energy),
BaseUnit('k', 'kayser', 'wavenumber', 100., '', 'k', d_length**-1),
BaseUnit('D', 'debye', 'electric dipole moment', 1.e-21/299792458., '', 'D',
         d_charge * d_length),
BaseUnit('hbar', 'hbar', 'angular momentum', 1.05457148e-34, '', r'\hbar',
         Dimensions(L=2, M=1, T=-1)),
BaseUnit('e', 'electron charge', 'charge', 1.602176565e-19, '', 'e', d_charge),
]),

('Angular units', [
BaseUnit('deg', 'degree', 'angle', 0.017453292519943295, '', 'deg',
         d_dimensionless),
BaseUnit('arcmin', 'arcminute', 'angle', 2.908882086657216e-4, '', 'arcmin',
         d_dimensionless),
# NB we can't allow 'as' for arcseconds because of ambiguity with attoseconds
BaseUnit('asec', 'arcsecond', 'angle', 4.84813681109536e-6, '', 'asec',
         d_dimensionless),
]),

('Non-SI energy units', [
BaseUnit('eV', 'electron volt', 'energy',  1.602176487e-19, '', 'eV',
         d_energy),
BaseUnit('E_h', 'hartree', 'energy', 4.35974394e-18, '', 'E_h', d_energy),
BaseUnit('cal', 'thermodynamic calorie', 'energy', 4.184, '', 'cal', d_energy),
BaseUnit('Ry', 'rydberg', 'energy', 13.60569253 * 1.602176487e-19, '', 'Ry',
         d_energy),
]),

('Non-SI mass units', [
BaseUnit('u', 'atomic mass unit', 'mass', 1.660538921e-27, '', 'u', d_mass),
BaseUnit('amu', 'atomic mass unit', 'mass', 1.660538921e-27, '', 'am', d_mass),
BaseUnit('Da', 'dalton', 'mass', 1.660538921e-27, '', 'Da', d_mass),
BaseUnit('m_e', 'electron mass', 'mass', 9.10938291e-31, '', 'm_e', d_mass),
]),

('Non-SI units of length, area and volume', [
# Non-SI length units
BaseUnit('Å', 'angstrom', 'length', 1.e-10, '', r'\AA', d_length),
BaseUnit('a0', 'bohr', 'length', 5.2917721092e-11, '', 'a_0', d_length),
# Non-SI area units
BaseUnit('b', 'barn', 'area', 1.e-28, '', 'b', d_area),
# Non-SI volume units
BaseUnit('l', 'litre', 'volume', 1.e-3, '', 'l', d_volume),
BaseUnit('L', 'litre', 'volume', 1.e-3, '', 'L', d_volume),
]),

('Non-SI time units', [
BaseUnit('min', 'minute', 'time', 60., '', 'min', d_time),
BaseUnit('hr', 'hour', 'time', 3600., '', 'hr', d_time),
BaseUnit('h', 'hour', 'time', 3600., '', 'h', d_time),
BaseUnit('dy', 'day', 'time', 86400., '', 'd', d_time),
BaseUnit('yr', 'year', 'time', 31557600., '', 'yr', d_time),
]),

('Astronomical units', [
BaseUnit('AU', 'astronomical unit', 'length', 1.495978707e11, '', 'AU',
         d_length),
BaseUnit('pc', 'parsec', 'length', 3.085677637634e16, '', 'pc', d_length),
BaseUnit('ly', 'light-year', 'length', 9.4607304725808e15, '', 'ly', d_length),
]),

('Imperial, customary and US units', [
# NB we can't use 'in' for inch because of a clash with min
BaseUnit('inch', 'inch', 'length', 0.0254, '', 'in', d_length),
BaseUnit('ft', 'foot', 'length', .3048, '', 'ft', d_length),
# NB we can't use 'yd' for yard because of a clash with yd (yoctodays!)
BaseUnit('yard', 'yard', 'length', .9144, '', 'yd', d_length),
BaseUnit('fur', 'furlong', 'length', 201.168, '', 'furlong', d_length),
BaseUnit('mi', 'mile', 'length', 1609.344, '', 'mi', d_length),
BaseUnit('gal', 'Imperial (UK) gallon', 'volume', 4.54609e-3, '', 'gal',
         d_volume),
BaseUnit('pt', 'Imperial (UK) pint', 'volume', 5.6826125e-4, '', 'pt',
         d_volume),
BaseUnit('USgal', 'US liquid gallon', 'volume', 3.785411783e-3, '', 'USgal',
         d_volume),
BaseUnit('USpt', 'US liquid pint', 'volume', 4.73176472875e-4, '', 'USpt',
         d_volume),
BaseUnit('st', 'stone', 'mass', 6.35029318, '', 'st', d_mass),
BaseUnit('lb', 'pound', 'mass', 0.45359237, '', 'lb', d_mass),
BaseUnit('oz', 'ounce', 'mass', 0.028349523125, '', 'oz', d_mass),
]),

('Maritime units', [
BaseUnit('NM', 'nautical mile', 'length', 1852., '', 'NM', d_length),
BaseUnit('kn', 'knot', 'speed', 1852., '', 'kn', d_length/d_time),
]),

('Silly units', [
BaseUnit('fir', 'fikin', 'mass', 40.8233133, '', 'fir', d_mass),
BaseUnit('ftn', 'fortnight', 'time', 1.2096e6, '', 'ftn', d_time),
]),

('Miscellaneous units', [
BaseUnit('Td', 'townsend', 'reduced electric field', 1.e-21, '', 'Td',
         d_voltage * d_area),
BaseUnit('Jy', 'jansky', 'spectral flux density', 1.e-26, '', 'Jy',
         d_energy / d_area)     # W.m-2.s-1
]),
)

# create a dictionary mapping the BaseUnit stems (as keys) to the BaseUnit
base_unit_stems = {}
for base_unit_group in base_units:
    for base_unit in base_unit_group[1]:
        base_unit_stems[base_unit.stem] = base_unit




# units.py
# A class representing the units of a physical quantity.
#
# Copyright (C) 2012-2016 Christian Hill
# Department of Physics and Astronomy, University College London
# christian.hill@ucl.ac.uk
#
# This file is part of PyQn
#
# PyQn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyQn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyQn.  If not, see <http://www.gnu.org/licenses/>

import copy
from .dimensions import Dimensions
from .dimensions import (d_dimensionless, d_length, d_energy, d_time,
                         d_temperature)
from .atom_unit import AtomUnit, UnitsError, feq

h, NA, c, kB = (6.62607015e-34, 6.02214076e+23, 299792458.0, 1.380649e-23)

class Units(object):
    """
    A class to represent the units of a physical quantity.

    """

    def __init__(self, units):
        """
        Initialize this Units from a list of AtomUnit objects or by parsing
        a string representing the units.

        """

        if type(units) is Units:
            self.__init__(units.atom_units)
        elif type(units) is str:
            self.__init__(self.parse(units).atom_units)
        elif type(units) is list:
            self.atom_units = copy.deepcopy(units)
        else:
            raise TypeError('Attempt to initialize Units object with'
                            ' argument units of type %s' % type(units))
        # also get the dimensions of the units
        self.dims = self.get_dims()

    def has_units(self):
        if self.atom_units:
            return True
        return False

    def get_dims(self):
        """
        Return a Dimensions object representing the dimensions of this
        Units.

        """

        dims = Dimensions()
        for atom_unit in self.atom_units:
            dims *= atom_unit.dims
        return dims

    @classmethod
    def parse(self, s_compoundunit):
        """
        Parse the string s_compoundunit and return the corresponding
        Units object.

        """

        div_fields = s_compoundunit.split('/')
        ndiv_fields = len(div_fields)
        compound_unit = Units.parse_mult_units(div_fields[0])
        for div_field in div_fields[1:]:
            compound_unit = compound_unit / Units.parse(div_field)
        return compound_unit

    @classmethod
    def parse_mult_units(self, munit):
        """
        Parse a string of units multiplied together (indicated by '.'),
        returning the corresponding Units object.

        """
        atom_units = []
        for s_unit in munit.split('.'):
            atom_unit = AtomUnit.parse(s_unit)
            if atom_unit.base_unit.stem != '1':
                # the unity 'unit' is not really a unit
                atom_units.append(atom_unit)
        return Units(atom_units)

    def _find_atom(self, atom_unit):
        """
        Return the index of atom_unit in the atom_units list, if it exists,
        ignoring any exponent. Otherwise return None.

        """

        for i,my_atom_unit in enumerate(self.atom_units):
            if my_atom_unit.prefix_base_eq(atom_unit):
                return i
        return None

    def __mul__(self, other):
        """ Return the product of this Units object with another. """

        if other == 1:
            return copy.deepcopy(self)

        if type(other) == str:
            other = Units(other)

        product = Units(self.atom_units)
        for other_atom_unit in other.atom_units:
            i = product._find_atom(other_atom_unit)
            if i is None:
                # new prefix and BaseUnit; append to the resulting product
                product.atom_units.append(other_atom_unit)
            else:
                # this prefix and BaseUnit is already in the AtomUnits
                # of the first operand: update its exponent in the product
                product.atom_units[i].exponent += other_atom_unit.exponent
                product.atom_units[i].dims *= other_atom_unit.dims
                if product.atom_units[i].exponent == 0:
                    # this AtomUnit has cancelled:
                    del product.atom_units[i]
        product.dims = product.get_dims()
        return product
    def __rmul__(self, other):
        if type(other) == str:
            other = Units(other)
        elif other == 1:
            other = Units('1')
        return self.__mul__(other)

    def __truediv__(self, other):
        """ Return the ratio of this Units divided by another. """
        if type(other) == str:
            other = Units(other)
        elif other == 1:
            other = Units('1')
        ratio = Units(self.atom_units)
        for other_atom_unit in other.atom_units:
            i = ratio._find_atom(other_atom_unit)
            if i is None:
                # new prefix and BaseUnit; append to the resulting ratio
                ratio.atom_units.append(other_atom_unit**-1)
            else:
                # this prefix and BaseUnit is already in the AtomUnits
                # of the first operand: update its exponent in the ratio
                ratio.atom_units[i].exponent -= other_atom_unit.exponent
                ratio.atom_units[i].dims /= other_atom_unit.dims
                if ratio.atom_units[i].exponent == 0:
                    # this AtomUnit has cancelled:
                    del ratio.atom_units[i]
        ratio.dims = ratio.get_dims()
        return ratio
        
    def __rdiv__(self, other):
        if type(other) == str:
            other = Units(other)
        elif other == 1:
            other = Units('1')
        return other.__truediv__(self)
 
    def __pow__(self, power):
        result_atom_units = []
        for atom_unit in self.atom_units:
            result_atom_units.append(atom_unit**power)
        return Units(result_atom_units)
        
    def __str__(self):
        """ String representation of this Units. """
        return '.'.join([str(atom_unit) for atom_unit in self.atom_units])

    __repr__ = __str__

    def __eq__(self, other):
        """ Test for equality with another Units object. """
        if other is None:
            return False
        elif other == 1:
            return self.get_dims() == d_dimensionless
        if self.get_dims() != other.get_dims():
        # obviously the units aren't the same if they have different dimensions
            return False
        # if two Units objects have the same dimensions, they are equal if
        # their conversion factors to SI units are the same: 
        if feq(self.to_si(), other.to_si()):
            return True
        return False

    def __ne__(self, other):
        return not self == other

    def to_si(self):
        """ Return the factor needed to convert this Units to SI. """
        fac = 1.
        for atom_unit in self.atom_units:
            fac *= atom_unit.si_fac
        return fac

    @property
    def html(self):
        h = []
        n = len(self.atom_units)
        for i,atom_unit in enumerate(self.atom_units):
            h.extend([atom_unit.prefix or '', atom_unit.base_unit.stem])
            if atom_unit.exponent != 1:
                h.append('<sup>{:d}</sup>'.format(atom_unit.exponent))
            if i < n-1:
                h.append(' ')
        return ''.join(h)


    @property
    def latex(self):
        e = []
        n = len(self.atom_units)
        for i, atom_unit in enumerate(self.atom_units):
            # TODO use proper LaTeX for prefix.
            e.extend([atom_unit.prefix or '', atom_unit.base_unit.latex])
            if atom_unit.exponent != 1:
                e.append('^{' + str(atom_unit.exponent) + '}')
            if i < n-1:
                e.append(r'\,')
        return r'\mathrm{' + ''.join(e) + '}'


    def conversion(self, other, force=None, strict=False):
        """
        Return the factor required to convert this Units to
        another. Their dimensions have to match, unless force is set
        to allow special cases:
        force='spec':   spectroscopic units: conversions between [L-1],
        [M,L2,T-2], [T-1] are allowed (i.e. cm-1, s-1, and J can be
        interconverted through factors of h, hc).
 
        """

        if type(other) == str:
            other = Units(other)
            
        conversion_method = {
                             'spec': self.spec_conversion,
                             'mol': self.mol_conversion,
                             'kBT': self.kBT_conversion
                             }

        self_dims, other_dims = self.get_dims(), other.get_dims()
        if self_dims != other_dims:
            try:
                return conversion_method[force](other)
            except KeyError:
                raise UnitsError('Failure in units conversion: units %s[%s] and'
                                 ' %s[%s] have different dimensions'
                           % (self, self.get_dims(), other, other.get_dims()))
        return self.to_si() / other.to_si()
    
    def kBT_conversion(self, other):
        from_dims = self.get_dims()
        to_dims = other.get_dims()
        fac = self.to_si()
        
        if from_dims == d_energy and to_dims == d_temperature:
            fac = fac / kB
        elif from_dims == d_temperature and to_dims == d_energy:
            fac = fac * kB
        else:
            raise UnitsError('Failure in conversion of units: was expecting to '
                             'covert between energy and temperature')
        return fac/other.to_si()

    def mol_conversion(self, other):
        from_dims = self.get_dims() #original unit dimensions
        to_dims = other.get_dims()  #desired unit dimensions
        fac = self.to_si()          #factor needed to conver to SI units
        
        if from_dims.dims[4] == to_dims.dims[4]:
            raise UnitsError('Failure in conversion of units: no '
                             'different in quantity dimensions between %s and %s'
                             % from_dims, to_dims)
        elif from_dims.dims[4] > to_dims.dims[4]:
            fac = fac/(NA**(from_dims.dims[4]-to_dims.dims[4]))
        else:
            fac = fac*(NA**(to_dims.dims[4]-from_dims.dims[4]))
        return fac/other.to_si()
        
    def spec_conversion(self, other):
        d_wavenumber = d_length**-1
        d_frequency = d_time**-1
        d_wavelength = d_length
        
        from_dims = self.get_dims()
        to_dims = other.get_dims()
        fac = self.to_si()
        if from_dims == d_wavenumber:
            fac *= h*c
        elif from_dims == d_frequency:
            fac *= h
        elif from_dims != d_energy:
            raise UnitsError('Failure in conversion of spectroscopic units:'
                ' I only recognise from-units of wavenumber, energy and'
                ' frequency but got %s' % str(self))
        if to_dims == d_wavenumber:
            fac /= h*c
        elif to_dims == d_frequency:
            fac /= h
        elif to_dims != d_energy:
            raise UnitsError('Failure in conversion of spectroscopic units:'
                ' I only recognise to-units of wavenumber, energy and'
                ' frequency but got %s' % str(other))
        return fac / other.to_si()

def convert(from_units, to_units):
    return Units(from_units).conversion(to_units)
